import numpy as np
import matplotlib.pyplot as plt


def plot_points(collocation_points, vertice_points):
    figure, axis = plt.subplots(2, 1)
    axis[0].invert_yaxis()
    axis[0].scatter(vertice_points[:,1], vertice_points[:,0], label='vertices')
    axis[0].scatter(collocation_points[:,1], collocation_points[:,0], label='collocation points')
    axis[0].set_title("X x Y")
    axis[0].legend(loc='best')

    # For Cosine Function
    axis[1].scatter(vertice_points[:,1], vertice_points[:,2], label='vertices')
    axis[1].scatter(collocation_points[:,1], collocation_points[:,2], label='collocation points')
    axis[1].set_title("Z x Y")
    axis[1].legend(loc='best')
    plt.show()


def get_span_y_distr(n, span_partition, distribution_type):
    vertice_points = np.zeros(n + 1)
    collocation_points = np.zeros(n)

    vertice_points[0] = span_partition * 1 / 2 * \
        (1 - np.cos(0))  # O primeiro ponto é 0
    for i in range(n):
        # Componente Y dos pontos
        if distribution_type == 'cosine':
            vertice_points[i+1] = span_partition * 1 / \
                2 * (1 - np.cos(((i+1) * np.pi / n)))
            collocation_points[i] = span_partition * 1 / 2 * \
                (1 - np.cos(((i+1) * np.pi / n) - (np.pi / (2 * n))))
        elif distribution_type == 'linear':
            vertice_points[i+1] = span_partition * ((i+1)/n)
            collocation_points[i] = span_partition * ((i)/n + (i+1)/n) * 1 / 2
        else:
            raise Exception(
                "Invalid 'distribution_type' value. Expected 'cosine' or 'linear'.")

    span_y_distribution = {
        'vertice_points': vertice_points,
        'collocation_points': collocation_points
    }
    return span_y_distribution


def get_span_x_distr(cp_y_component, vp_y_component, chord_i, chord_ii, offset_i, offset_ii, span_partition):
    n = len(cp_y_component)
    vp_x_component = np.zeros(n+1)
    cp_x_component = np.zeros(n)

    vp_x_component[0] = \
    0.25*(chord_ii - chord_i) / span_partition * vp_y_component[0] + 0.25 * chord_i + \
    (offset_ii - offset_i) / span_partition * vp_y_component[0] + offset_i
    for i in range(n):
        vp_x_component[i+1] = \
        0.25*(chord_ii - chord_i) / span_partition * vp_y_component[i+1] + 0.25 * chord_i + \
        (offset_ii - offset_i) / span_partition * vp_y_component[i+1] + offset_i

        cp_x_component[i] = \
        0.25*(chord_ii - chord_i) / span_partition * cp_y_component[i] + 0.25 * chord_i + \
        (offset_ii - offset_i) / span_partition * cp_y_component[i] + offset_i
    
    span_x_distribution = {
        'vertice_points': vp_x_component,
        'collocation_points': cp_x_component
    }
    return span_x_distribution


def get_span_z_distr(cp_y_component, vp_y_component, dihedral_partition):
    n = len(cp_y_component)
    vp_z_component = np.zeros(n+1)
    cp_z_component = np.zeros(n) 

    vp_z_component[0] = np.tan(dihedral_partition) * vp_y_component[0]
    for i in range(n):
        vp_z_component[i+1] = np.tan(dihedral_partition) * vp_y_component[i+1]
        cp_z_component[i] = np.tan(dihedral_partition) * cp_y_component[i]
    
    span_z_distribution = {
        'vertice_points': vp_z_component,
        'collocation_points': cp_z_component
    }
    return span_z_distribution


def get_twist_distr(cp_y_component, twist_i, twist_ii, span_partition):
    n = len(cp_y_component)
    span_twist_distribution = np.zeros(n)
    for i in range(n):
        span_twist_distribution[i] = (twist_ii - twist_i) / span_partition * cp_y_component[i] + twist_i
    return span_twist_distribution


def get_local_chord(y, chord_i, chord_ii, span_partition):
    return (chord_ii - chord_i) / span_partition * y + chord_i

def get_euler_matrix(dihedral, twist, sweep):
    # Utilizando-se rotação XYZ -> Ângulos de Tait-Bryan
    # REFERÊNCIA: https://en.wikipedia.org/wiki/Euler_angles#Conversion_to_other_orientation_representations
    c1 = np.cos(dihedral)
    c2 = np.cos(twist)
    c3 = np.cos(sweep)
    s1 = np.sin(dihedral)
    s2 = np.sin(twist)
    s3 = np.sin(sweep)
    euler_matrix = np.array([
        [c2*c3, -c2*s3, s2],
        [c1*s3+c3*s1*s2, c1*c3-s1*s2*s3, -c2*s1],
        [s1*s3-c1*c3*s2, c3*s1+c1*s2*s3, c1*c2]
        ])
    return euler_matrix



import numpy as np
import math
from typing import List, Union, Optional
from dataclasses import dataclass, field
from lltdrek.models.flight_condition import FlightCondition
import lltdrek.utils.geometry as geo


@dataclass(repr=False, eq=False, match_args=False, slots=True)
class Wing:
    spans: List[float]
    chords: List[float]
    offsets: List[float]
    twist_angles: List[float]
    dihedral_angles: List[float]
    airfoils: List[str]
    surface_name: str
    N_panels: Union[int, List[int]]
    x_pos: float = 0
    z_pos: float = 0
    distribution_type: str = 'linear'
    sweep_check: bool = False
    # Properties obtained through get_mesh method
    total_span: float = field(init=False)
    span_panel_numbers: List[int] = field(init=False)
    collocation_points: np.ndarray = field(init=False)
    vertice_points: np.ndarray = field(init=False)
    u_a: np.ndarray = field(init=False)
    u_n: np.ndarray = field(init=False)
    u_s: np.ndarray = field(init=False)
    cp_lengths: np.ndarray = field(init=False)
    cp_dsl: np.ndarray = field(init=False)
    cp_areas: np.ndarray = field(init=False)
    cp_chords: np.ndarray = field(init=False)
    cp_macs: np.ndarray = field(init=False)
    cp_reynolds: np.ndarray = field(init=False)
    cp_airfoils: List = field(init=False)
    airfoil_data: dict = field(init=False)
    MAC: float = field(init=False)
    partition_areas: np.ndarray = field(init=False)
    total_area: float = field(init=False)
    AR: float = field(init=False)
    parent_wing: Optional[str] = field(init=False)

    allowed_distribution_types = [
        "linear",
        "cosine"
    ]

    def __post_init__(self) -> None:
        if self.distribution_type not in self.allowed_distribution_types:
            raise ValueError("Invalid 'distribution_type' value.")
        # Convert degree to rad
        self.twist_angles = [angle * np.pi / 180 for angle in self.twist_angles]
        self.dihedral_angles = [angle * np.pi / 180 for angle in self.dihedral_angles]
        self.parent_wing = None # Usado para asas espelhadas
    

    def __repr__(self) -> None:
        return self.surface_name


    def generate_mesh(self) -> None:
        self.total_span = sum(self.spans)
        # Number of panels of each partition
        self.span_panel_numbers = [value / self.total_span * self.N_panels for value in self.spans]
        self.span_panel_numbers = [math.ceil(int(i)) for i in self.span_panel_numbers]
        # Add a panel to compensate rounding
        if sum(self.span_panel_numbers) == self.N_panels - 1:
            self.span_panel_numbers[-1] += 1

        self.partition_areas = np.zeros(len(self.spans))
        
        self.collocation_points = np.zeros([self.N_panels, 3])
        self.vertice_points = np.zeros([self.N_panels + 1, 3])

        self.u_a = np.zeros([self.N_panels, 3]) # Vetor unitário colinear à corda
        self.u_n = np.zeros([self.N_panels, 3]) # Vetor unitário normal à corda
        self.u_s = np.zeros([self.N_panels, 3]) # Vetor perpendicular ao plano do perfil

        self.cp_lengths = np.zeros([self.N_panels, 3]) # Vetor de comprimento de cada painel
        self.cp_dsl = np.zeros([self.N_panels, 3]) # Vetor de comprimento ao longo da envergadura adimensional (dimensionless spanwise length vector)

        self.cp_areas = np.zeros(self.N_panels) # Área de cada painel
        self.cp_chords = np.zeros(self.N_panels) # Distribuição de cordas dos pontos de colocação
        self.cp_macs = np.zeros(self.N_panels) # Corda média aerodinâmica de cada painel

        # TODO: explain what these variables mean
        idx_n = 0
        span_incremental = 0
        height_incremental = 0
        MAC = 0
        for i, span_partition in enumerate(self.spans):
            n = self.span_panel_numbers[i]
            chord_i = self.chords[i]
            chord_ii = self.chords[i+1]
            offset_i = self.offsets[i]
            offset_ii = self.offsets[i+1]
            twist_i = self.twist_angles[i]
            twist_ii = self.twist_angles[i+1]
            dihedral_partition = self.dihedral_angles[i]

            # Ângulo de enflechamento [rad] da partição atual
            sweep_partition = np.arctan( (0.25*chord_ii - 0.25*chord_i + offset_i)/span_partition ) if self.sweep_check is True else 0

            span_y_distr = geo.get_span_y_distr(n, span_partition, self.distribution_type)
            cp_y_component = span_y_distr['collocation_points']
            vp_y_component = span_y_distr['vertice_points']

            span_x_distr = geo.get_span_x_distr(cp_y_component,vp_y_component,chord_i,chord_ii,offset_i,offset_ii,span_partition)
            cp_x_component = span_x_distr['collocation_points']
            vp_x_component = span_x_distr['vertice_points']

            span_z_distr = geo.get_span_z_distr(cp_y_component,vp_y_component,dihedral_partition)
            cp_z_component = span_z_distr['collocation_points']
            vp_z_component = span_z_distr['vertice_points']

            cp_twist_distr = geo.get_twist_distr(cp_y_component, twist_i, twist_ii, span_partition)

            # Alocar o primeiro ponto em vertice_points
            if i == 0:
                self.vertice_points[0][0] = vp_x_component[0]
                self.vertice_points[0][1] = vp_y_component[0]
                self.vertice_points[0][2] = vp_z_component[0]

            for j, _ in enumerate(cp_y_component):
                # Componente Y dos CP / VP
                self.vertice_points[idx_n+j+1][1] = span_incremental + vp_y_component[j+1]
                self.collocation_points[idx_n+j][1] = span_incremental + cp_y_component[j]
                # Componente X
                self.vertice_points[idx_n+j+1][0] = vp_x_component[j+1]
                self.collocation_points[idx_n+j][0] = cp_x_component[j]
                # Componente Z
                self.vertice_points[idx_n+j+1][2] = height_incremental + vp_z_component[j+1]
                self.collocation_points[idx_n+j][2] = height_incremental + cp_z_component[j]
                # Ângulo de torção (twist geométrico) do ponto de colocação [testar]
                twist_cp = cp_twist_distr[j]

                # Vetores de posicionamento do painel (colinear e perpendicular à corda)
                euler_matrix = geo.get_euler_matrix(dihedral_partition, twist_cp, sweep_partition)
                self.u_a[idx_n+j] = euler_matrix.dot(np.array([1, 0, 0]))
                self.u_n[idx_n+j] = euler_matrix.dot(np.array([0, 0, 1]))
                self.u_s[idx_n+j] = np.cross(self.u_a[idx_n+j], self.u_n[idx_n+j])

                chord_vp_j = geo.get_local_chord(vp_y_component[j], chord_i, chord_ii, span_partition)
                chord_vp_jj = geo.get_local_chord(vp_y_component[j+1], chord_i, chord_ii, span_partition)
                chord_cp = geo.get_local_chord(cp_y_component[j], chord_i, chord_ii, span_partition)
                self.cp_chords[idx_n+j] = chord_cp

                # Corda média aerodinâmica  do painel
                cp_mac = (2/3)*(chord_vp_j ** 2 + chord_vp_j * chord_vp_jj + chord_vp_jj ** 2)/(chord_vp_j + chord_vp_jj)
                self.cp_macs[idx_n+j] = cp_mac
                for k in range(3):
                    self.cp_lengths[idx_n+j][k] = self.vertice_points[idx_n+j+1][k] - self.vertice_points[idx_n+j][k]

                # Vetor comprimento do painel
                cp_length_x = self.cp_lengths[idx_n+j][0]
                cp_length_y = self.cp_lengths[idx_n+j][1]
                cp_length_z = self.cp_lengths[idx_n+j][2]

                # Área do painel
                cp_area = 0.5*(chord_vp_j + chord_vp_jj)*math.sqrt( cp_length_y ** 2 + cp_length_z ** 2 )
                self.cp_areas[idx_n+j] = cp_area
                self.partition_areas[i] += cp_area

                # Cálculo do dimensionless spanwise length
                for k in range(3): 
                    self.cp_dsl[idx_n+j][k] = (cp_mac * self.cp_lengths[idx_n+j][k]) / cp_area

            # Razão de afilamento da partição atual
            partition_lambda = chord_ii / chord_i
            # Corda média aerodinâmica da asa
            MAC += self.partition_areas[i]*(2/3)*chord_i*(1+partition_lambda+partition_lambda**2)/(1+partition_lambda)
            idx_n += n
            span_incremental += span_partition
            height_incremental += vp_z_component[-1]
    
        # Add coordinates into cp's and vp's
        self.collocation_points[:,0] += self.x_pos
        self.vertice_points[:,0] += self.x_pos
        self.collocation_points[:,2] += self.z_pos
        self.vertice_points[:,2] += self.z_pos
        
        # Global values
        self.total_area = sum(self.partition_areas)
        self.MAC = MAC / sum(self.partition_areas)
        self.AR = (2 * self.total_span) ** 2 / (2 * sum(self.partition_areas))


    def setup_airfoil_data(self, flight_condition: FlightCondition, airfoils_data_dict: dict) -> None:
        """
        TODO: Explicar este método e suas variáveis
        """
        self.cp_reynolds = np.zeros(self.N_panels)
        self.cp_airfoils = []
        idx_n = 0
        for i, span_partition in enumerate(self.spans):
            n = self.span_panel_numbers[i]
            airfoil_i = self.airfoils[i]
            airfoil_ii = self.airfoils[i+1]
            for j in range(self.span_panel_numbers[i]):
                if airfoil_i == airfoil_ii:
                    self.cp_airfoils.append([1, [airfoil_i]])
                else:
                    merge_parameter = (self.collocation_points[idx_n+j][1]-self.vertice_points[idx_n][1]) / span_partition
                    self.cp_airfoils.append([merge_parameter, [airfoil_i, airfoil_ii]])
            idx_n += n

        for i, panel_chord in enumerate(self.cp_chords):
            reynolds_number = panel_chord * flight_condition.V_inf / flight_condition.nu
            self.cp_reynolds[i] = reynolds_number
        
        unique_airfoils = list(set(self.airfoils))
        self.airfoil_data = {}
        for airfoil, airfoil_data in airfoils_data_dict.items():
            if airfoil in unique_airfoils:
                self.airfoil_data[airfoil] = airfoil_data



from typing import Dict, List
from dataclasses import dataclass, field
import numpy as np
import copy
from lltdrek.models.wing import Wing
from lltdrek.models.flight_condition import FlightCondition
from lltdrek.utils import velocity
from lltdrek.utils.timeit import timeit
from lltdrek.utils.lookup import get_airfoil_data


"""
ind_velocities_list = [ind_velocities_dict_1, ind_velocities_dict_2, ...]

ind_velocities_dict = {
    "surface_name": {
        "surface_name": [[velocity_distribution_panel_1], [velocity_distribution_panel_2], ...]
    }
}

G_dict = {
    "surface_name": "G_array",
}

total_velocity_dict = {
   "surface_name": "total_velocity"
}

aoa_eff_dict = {
    "surface_name": "aoa_eff_list",
}
"""


@dataclass(repr=False, eq=False, match_args=False, slots=True)
class WingPool:
    wing_list: List[Wing]
    flight_condition: FlightCondition
    initial_G: Dict = None
    S_ref: float = None
    c_ref: float = None
    ind_velocities_list: List = field(init=False)
    G_dict: Dict = field(init=False)
    complete_wing_pool: List[Wing] = field(init=False)
    total_panels: int = field(init=False)

    # @timeit
    def __post_init__(self):
        self.G_dict = {}
        self.complete_wing_pool = []
        self.ind_velocities_list = []
        self.total_panels = 0
        
        if self.c_ref is None:
            self.c_ref = self.wing_list[0].MAC
        
        if(len(set(self.wing_list)) == len(self.wing_list)):
            self.build_complete_wing_pool()
        else:
            raise Exception("As asas que compõem a wing pool precisam ter nomes únicos")
        
        if self.S_ref is None:
            self.S_ref = 0
            for _, wing in enumerate(self.complete_wing_pool):
                self.S_ref += wing.total_area

        for _, wing in enumerate(self.complete_wing_pool):
            self.total_panels += wing.N_panels
            if self.initial_G == None:
                self.G_dict[wing.surface_name] = [0.1 for _ in range(wing.N_panels)]
            else:
                for surface_name, G_list in self.initial_G.items():
                    self.G_dict[surface_name] = G_list
                    self.G_dict[surface_name+"_mirrored"] = G_list
        
        for v_inf_array in self.flight_condition.v_inf_list:
            ind_velocities_dict = self.calculate_induced_velocities(v_inf_array)
            self.ind_velocities_list.append(ind_velocities_dict)


    def build_complete_wing_pool(self):
        """
        Method that builds a wing pool with mirrored wing objects.
        This is the list that will be used in calculations
        """
        for wing in self.wing_list:
            mirrored_wing = copy.deepcopy(wing)
            mirrored_wing.surface_name += "_mirrored"

            # Mirror y-coordinate of required values
            mirrored_wing.u_a[:,1] = -1 * mirrored_wing.u_a[:,1]
            mirrored_wing.u_n[:,1] = -1 * mirrored_wing.u_n[:,1]
            mirrored_wing.u_s[:,0] = -1 * mirrored_wing.u_s[:,0]
            mirrored_wing.u_s[:,2] = -1 * mirrored_wing.u_s[:,2]
            mirrored_wing.collocation_points[:,1] = -1 * mirrored_wing.collocation_points[:,1]
            mirrored_wing.vertice_points[:,1] = -1 * mirrored_wing.vertice_points[:,1]
            mirrored_wing.cp_lengths[:,1] = -1 * mirrored_wing.cp_lengths[:,1]
            mirrored_wing.cp_dsl[:,0] = -1 * mirrored_wing.cp_dsl[:,0]
            mirrored_wing.cp_dsl[:,2] = -1 * mirrored_wing.cp_dsl[:,2]
            
            mirrored_wing.parent_wing = wing.surface_name

            self.complete_wing_pool.append(wing)
            self.complete_wing_pool.append(mirrored_wing)


    def update_solution(self, G_solution) -> None:
        """
        Method that splits the G_solution list, obtained by solving the 
        the main equations, into separate solution lists for each wing in the
        complete_wing_pool
        """
        G_dict = {}
        global_counter = 0
        for wing in self.complete_wing_pool:
            counter = 0
            G_array = np.zeros(wing.N_panels)
            while counter < wing.N_panels:
                G_array[counter] = G_solution[global_counter]
                counter += 1
                global_counter += 1
            G_dict[wing.surface_name] = G_array
        return G_dict
    
    # @timeit
    def calculate_induced_velocities(self, v_inf_array: np.ndarray) -> dict:
        # Essa função calcula tudo para um angulo de ataque e é chamada para cada aoa no __post_init__
        ind_velocities_dict = {}
        for wing in self.complete_wing_pool:
            ind_velocities_dict[wing.surface_name] = {}
        
        for wing_cp in self.complete_wing_pool:
            collocation_points = wing_cp.collocation_points
            cp_macs = wing_cp.cp_macs
            for wing_vp in self.complete_wing_pool:
                vertice_points = wing_vp.vertice_points
                v_ij_distr = velocity.get_induced_velocity_distribution(
                    collocation_points, cp_macs, vertice_points, v_inf_array, wing_vp.surface_name, self.flight_condition.ground_effect_check, self.flight_condition.h
                )
                ind_velocities_dict[wing_cp.surface_name][wing_vp.surface_name] = v_ij_distr
        return ind_velocities_dict


    def calculate_total_velocity(self, aoa_idx: int, G_dict: dict):
        """
        WIP
        Method that calculates the sum of vij * G  + v_inf of all wings
        - Dá pra acelerar esse método calculando somente a velocidade dos objetos originais e copiando para os objetos 
        espelhados
        """
        v_inf_array = self.flight_condition.v_inf_list[aoa_idx]
        
        ind_velocities_dict = self.ind_velocities_list[aoa_idx]
        total_velocity_dict = {}
        for wing in self.complete_wing_pool:
            total_velocity_dict[wing.surface_name] = np.zeros((wing.N_panels,3))

        for wing_i in self.complete_wing_pool:
            for i, _ in enumerate(total_velocity_dict[wing_i.surface_name]):
                total_velocity_dict[wing_i.surface_name][i] += v_inf_array
                for wing_j in self.complete_wing_pool:
                    G = G_dict[wing_j.surface_name]
                    ind_velocities_distr = ind_velocities_dict[wing_i.surface_name][wing_j.surface_name][i]
                    for j, v_ij in enumerate(ind_velocities_distr):
                        total_velocity_dict[wing_i.surface_name][i] += v_ij * G[j]
        return total_velocity_dict


    def calculate_aoa_eff(self, total_velocity_dict: dict) -> dict:
        """
        calcular o angulo de ataque efetivo de cada painel de cada asa
        """
        aoa_eff_dict = {}
        for wing in self.wing_list:
            total_velocity_distr = total_velocity_dict[wing.surface_name]
            aoa_eff_distr = np.zeros(wing.N_panels)
            for idx, velocity in enumerate(total_velocity_distr):
                aoa_eff_distr[idx] = np.arctan(np.dot(velocity, wing.u_n[idx]) / np.dot(velocity, wing.u_a[idx]))

            aoa_eff_dict[wing.surface_name] = aoa_eff_distr
            aoa_eff_dict[wing.surface_name+"_mirrored"] = aoa_eff_distr
        return aoa_eff_dict

    
    def get_Cl_distribution(self, aoa_eff_dict: dict, cl_alpha_check: bool) -> dict:
        """
        O método pode retornar tanto a distribuição de Cl quanto a distribuição de Clalpha
        do sistema de asas
        Não é utilizado atualmente
        """
        Cl_dict = {}
        for wing in self.wing_list:
            Cl_distr = np.zeros(wing.N_panels)
            aoa_eff_distr = aoa_eff_dict[wing.surface_name]
            for i, _ in enumerate(wing.collocation_points):
                Cl_distr[i] = get_airfoil_data(
                    wing.cp_airfoils[i],
                    wing.cp_reynolds[i],
                    aoa_eff_distr[i] * 180 / np.pi,
                    wing.airfoil_data,
                    cl_alpha_check = cl_alpha_check
                )
            Cl_dict[wing.surface_name] = Cl_distr
        return Cl_dict


import numpy as np


def get_airfoil_data(
    cp_airfoil: list,
    cp_reynolds: float,
    aoa: float,
    airfoil_data: dict,
    cl_alpha_check: bool
) -> float:
    """
    - Função que retorna os dados de perfil do painel de uma asa.
    - Retorna o Cl, Cl_alfa (não linear) bem como dados lineares (Cl_alfa, Cl0, Cm0)
    de um painel
    - A função verifica se o perfil do painel (cp_airfoil) é uma mescla
    entre os dois perfis das extremidades da envergadura e realiza uma interpolação
    de Cl's caso necessário.
    - Para o caso de um perfil com mescla, a interpolação de dados se dá
    pela seguinte formula: 
        - Cl_perfil = Cl_raiz * (1 - merge_parameter) + Cl_ponta * (1 - merge_parameter),
    onde merge_parameter = cp_airfoil[0]
    - A verificação da mescla de dados entre perfis é feita olhando-se o tamanho
    do segundo índice da lista cp_airfoil (caso a lista de perfis seja > 1)
        - Caso len(cp_airfoil[1]) > 1, ocorrerá uma mescla de dados
    - TODO: Passar a distribuiçao de aoa's de uma asa e retornar todos os cl's / cl_alpha's
    """
    # Verificar se o painel em questão é uma mescla de perfis
    if len(cp_airfoil[1]) > 1:
        merge_parameter = cp_airfoil[0]

        airfoil_root = cp_airfoil[1][0]
        Cl_root = cl_lookup(airfoil_data[airfoil_root], cp_reynolds, aoa, cl_alpha_check, linear_check=False)
        airfoil_tip = cp_airfoil[1][1]
        Cl_tip = cl_lookup(airfoil_data[airfoil_tip], cp_reynolds, aoa, cl_alpha_check, linear_check=False)

        Cl = Cl_root * (1 - merge_parameter) + Cl_tip * merge_parameter
    else:
        airfoil = cp_airfoil[1][0]
        Cl = cl_lookup(airfoil_data[airfoil], cp_reynolds, aoa, cl_alpha_check, linear_check=False)
    return Cl


def get_linear_data_and_clmax(
    cp_airfoil: list,
    cp_reynolds: float,
    airfoil_data: dict
    ) -> dict:
    """
    Função que realizar o lookup nos dados lineares e clmax
    Número de Reynolds utilizado é o mais próximo de cp_reynolds
    """
    if len(cp_airfoil[1]) > 1:
        merge_parameter = cp_airfoil[0]

        airfoil_root = cp_airfoil[1][0]
        lookup_data_root = cl_lookup(airfoil_data[airfoil_root], cp_reynolds, aoa=None, cl_alpha_check=False, linear_check=True)
        airfoil_tip = cp_airfoil[1][1]
        lookup_data_tip = cl_lookup(airfoil_data[airfoil_tip], cp_reynolds, aoa=None, cl_alpha_check=False, linear_check=True)

        lookup_data = {
            "cl_alpha": lookup_data_root["cl_alpha"] * (1 - merge_parameter) + lookup_data_tip["cl_alpha"] * merge_parameter,
            "cl0": lookup_data_root["cl0"] * (1 - merge_parameter) + lookup_data_tip["cl0"] * merge_parameter,
            "cm0": lookup_data_root["cm0"] * (1 - merge_parameter) + lookup_data_tip["cm0"] * merge_parameter,
            "clmax": lookup_data_root["clmax"] * (1 - merge_parameter) + lookup_data_tip["clmax"] * merge_parameter,
        }
    else:
        airfoil = cp_airfoil[1][0]
        lookup_data = cl_lookup(airfoil_data[airfoil], cp_reynolds, aoa=None, cl_alpha_check=False, linear_check=True)
    
    return lookup_data


def cl_lookup(airfoil_data_dict: dict, reynolds_number: float, aoa: float, cl_alpha_check: bool, linear_check: bool) -> float:
    """
    - Função que realiza o lookup do numero de reynolds para, então, realizar o lookup de 
    aoa
    - Esta função retorna tanto um cl (aoa_list_lookup) quanto um cl_alpha (get_non_linear_cl_alpha)
    - A variável cl_alpha_check realiza o controle para chamar a função apropriada
    - A variável linear_check faz com que a função retorne dados lineares ou não
    """
    reynolds_list_str = list(airfoil_data_dict)
    reynolds_list = [float(reynolds) for reynolds in reynolds_list_str]
    if reynolds_number <= reynolds_list[0]:
        if linear_check:
            return {
                "cl_alpha": airfoil_data_dict[reynolds_list_str[0]]["cl_alpha"],
                "cl0": airfoil_data_dict[reynolds_list_str[0]]["cl0"],
                "cm0": airfoil_data_dict[reynolds_list_str[0]]["cm0"], 
                "clmax": airfoil_data_dict[reynolds_list_str[0]]["clmax"], 
            }

        cl_data = airfoil_data_dict[reynolds_list_str[0]]["cl_list"]
        cl = aoa_list_lookup(cl_data, aoa) if cl_alpha_check is False \
            else get_non_linear_cl_alpha(cl_data, aoa)
    elif reynolds_number >= reynolds_list[-1]:
        if linear_check:
            return {
                "cl_alpha": airfoil_data_dict[reynolds_list_str[-1]]["cl_alpha"],
                "cl0": airfoil_data_dict[reynolds_list_str[-1]]["cl0"],
                "cm0": airfoil_data_dict[reynolds_list_str[-1]]["cm0"], 
                "clmax": airfoil_data_dict[reynolds_list_str[-1]]["clmax"],
            }

        cl_data = airfoil_data_dict[reynolds_list_str[-1]]["cl_list"]
        cl = aoa_list_lookup(cl_data, aoa) if cl_alpha_check is False \
            else get_non_linear_cl_alpha(cl_data, aoa)
    else:
        for i in range(len(reynolds_list)-1):
            re_i = reynolds_list[i]
            re_ii = reynolds_list[i+1]
            if re_i <= reynolds_number <= re_ii:
                if linear_check:
                    cl_alpha_i = airfoil_data_dict[reynolds_list_str[i]]["cl_alpha"]
                    cl_alpha_ii = airfoil_data_dict[reynolds_list_str[i+1]]["cl_alpha"]
                    cl0_i = airfoil_data_dict[reynolds_list_str[i]]["cl0"]
                    cl0_ii = airfoil_data_dict[reynolds_list_str[i+1]]["cl0"]
                    cm0_i = airfoil_data_dict[reynolds_list_str[i]]["cm0"]
                    cm0_ii = airfoil_data_dict[reynolds_list_str[i+1]]["cm0"]
                    clmax_i = airfoil_data_dict[reynolds_list_str[i]]["clmax"]
                    clmax_ii = airfoil_data_dict[reynolds_list_str[i+1]]["clmax"]
                    return {
                        "cl_alpha": (cl_alpha_ii - cl_alpha_i)/(re_ii - re_i) * (reynolds_number - re_i) + cl_alpha_i,
                        "cl0": (cl0_ii - cl0_i)/(re_ii - re_i) * (reynolds_number - re_i) + cl0_i,
                        "cm0": (cm0_ii - cm0_i)/(re_ii - re_i) * (reynolds_number - re_i) + cm0_i,
                        "clmax": (clmax_ii - clmax_i)/(re_ii - re_i) * (reynolds_number - re_i) + clmax_i,
                    }

                cl_data_i = airfoil_data_dict[reynolds_list_str[i]]["cl_list"]
                cl_data_ii = airfoil_data_dict[reynolds_list_str[i+1]]["cl_list"]

                cl_i = aoa_list_lookup(cl_data_i, aoa) if cl_alpha_check is False \
                    else get_non_linear_cl_alpha(cl_data_i, aoa)
                cl_ii = aoa_list_lookup(cl_data_ii, aoa) if cl_alpha_check is False \
                    else get_non_linear_cl_alpha(cl_data_ii, aoa)
                
                cl = (cl_ii - cl_i)/(re_ii - re_i) * (reynolds_number - re_i) + cl_i
                break
            
    return cl


def aoa_list_lookup(cl_data: np.ndarray, aoa: float) -> float:
    """
    TODO: melhorar o código quando o alfa tá fora dos limites da lista
    """
    if aoa <= cl_data[0][0]:
        aoa_i = cl_data[0][0]
        aoa_ii = cl_data[1][0]
        cl_i = cl_data[0][1]
        cl_ii = cl_data[1][1]
    elif aoa >= cl_data[-1][0]:
        aoa_i = cl_data[-2][0]
        aoa_ii = cl_data[-1][0]
        cl_i = cl_data[-2][1]
        cl_ii = cl_data[-1][1]
    else:
        for i, _ in enumerate(cl_data[0:-1,0]):
            if cl_data[i][0] < aoa < cl_data[i+1][0]:
                aoa_i = cl_data[i][0]
                aoa_ii = cl_data[i+1][0]
                cl_i = cl_data[i][1]
                cl_ii = cl_data[i+1][1]
    
    cl_interp = (cl_ii - cl_i) / (aoa_ii - aoa_i) * (aoa - aoa_i) + cl_i
    return cl_interp


def get_non_linear_cl_alpha(cl_data: np.ndarray, aoa: float) -> float:
    """
    - Função que calcula o cl alpha para um dado
    aoa utilizando a fórmula da diferença dividida finita
    - A função precisa de 3 pontos para calcular a derivada
        - Caso aoa esteja no limite inferior, utiliza-se a versão 
        progressiva da fórmula
        - Caso aoa esteja no limite superior, utiliza-se a versão
        regressiva da fórmula
        - Caso contrário utiliza-se a fórmula centrada de 2 pontos
    - Todas as fórmulas possuem erro O(h²)
    - Referência: Métodos numéricos para engenharia, capítulo 6

    """
    if aoa <= cl_data[0][0]:
        aoa_i = cl_data[0][0]
        aoa_ii = cl_data[1][0]
        cl_i = cl_data[0][1]
        cl_ii = cl_data[1][1]
    elif aoa >= cl_data[-1][0]:
        aoa_i = cl_data[-2][0]
        aoa_ii = cl_data[-1][0]
        cl_i = cl_data[-2][1]
        cl_ii = cl_data[-1][1]
    else:
        for i, _ in enumerate(cl_data[0:-1,0]):
            if cl_data[i][0] < aoa < cl_data[i+1][0]:
                aoa_i = cl_data[i][0]
                aoa_ii = cl_data[i+1][0]
                cl_i = cl_data[i][1]
                cl_ii = cl_data[i+1][1]
    
    cl_alpha = (cl_ii - cl_i) / (aoa_ii - aoa_i)
    return cl_alpha


# Função para encontrar o índice do valor mais próximo em uma array
def find_closest(arr, val): return np.abs(arr - val).argmin()



import numpy as np
import numpy.linalg as npla
from lltdrek.models.wingpool import WingPool
from lltdrek.utils.lookup import get_airfoil_data, get_linear_data_and_clmax
from lltdrek.utils.timeit import timeit


def calculate_main_equation_simplified(
    v_inf_array: np.ndarray,
    aoa_idx: int,
    wing_pool: WingPool,
    matrix_dim: int
) -> np.ndarray:
    A_matrix = np.zeros([matrix_dim, matrix_dim])
    B_matrix = np.zeros(matrix_dim)

    i_glob = 0
    Cl_alpha_array = np.zeros(matrix_dim) # Used for debugging purposes
    Cl0_array = np.zeros(matrix_dim) # Used for debugging purposes
    for wing_i in wing_pool.complete_wing_pool:
            for i, _ in enumerate(wing_i.collocation_points):
                linear_data = get_linear_data_and_clmax(wing_i.cp_airfoils[i], wing_i.cp_reynolds[i], wing_i.airfoil_data)
                Cl0_i = linear_data["cl0"]
                Cl_alpha_i = linear_data["cl_alpha"] * 180 / np.pi
                Cl0_array[i_glob] = Cl0_i
                Cl_alpha_array[i_glob] = Cl_alpha_i
                alpha_zero_lift = -Cl0_i/Cl_alpha_i

                A_matrix[i_glob][i_glob] = 2 * npla.norm(np.cross(v_inf_array, wing_i.cp_dsl[i]))
                B_matrix[i_glob] = Cl_alpha_i * (np.dot(v_inf_array, wing_i.u_n[i]) - alpha_zero_lift)

                j_glob = 0
                for wing_j in wing_pool.complete_wing_pool:
                    v_ij_distr = wing_pool.ind_velocities_list[aoa_idx][wing_i.surface_name][wing_j.surface_name]
                    for j, _ in enumerate(wing_j.collocation_points):
                        v_ij = v_ij_distr[i][j]
                        A_matrix[i_glob][j_glob] += -1 * Cl_alpha_i * np.dot(v_ij, wing_i.u_n[i])
                        j_glob += 1
                i_glob += 1
    G_solution = npla.solve(A_matrix, B_matrix)
    return G_solution


def calculate_main_equation(
    total_velocity_dict: dict,
    aoa_eff_dict: dict,
    G_dict: dict,
    wing_pool: WingPool,
    matrix_dim: int,
    linear_check: bool
) -> np.ndarray:
    """
    Main system of equations -
    F(G) = R
    """
    R_array = np.zeros(matrix_dim)
    i_glob = 0
    Cl_array = np.zeros(matrix_dim) # Used for debugging purposes
    for wing in wing_pool.complete_wing_pool:
        if "_mirrored" in wing.surface_name: 
            i_glob += wing.N_panels
            continue
        N_panels = wing.N_panels
        aoa_eff_distr = aoa_eff_dict[wing.surface_name]
        G_list = G_dict[wing.surface_name]
        total_velocity_distr = total_velocity_dict[wing.surface_name]
        for i, _ in enumerate(wing.collocation_points):
            if linear_check:
                linear_data = get_linear_data_and_clmax(wing.cp_airfoils[i], wing.cp_reynolds[i], wing.airfoil_data)
                Cl0_i = linear_data["cl0"]
                Cl_alpha_i = linear_data["cl_alpha"] * 180 / np.pi
                Cl_i = Cl_alpha_i * aoa_eff_distr[i] + Cl0_i
            else:
                Cl_i = get_airfoil_data(
                    wing.cp_airfoils[i],
                    wing.cp_reynolds[i],
                    aoa_eff_distr[i] * 180 / np.pi,
                    wing.airfoil_data,
                    cl_alpha_check = False
                )
            Cl_array[i_glob] = Cl_i
            R_array[i_glob] = 2 * npla.norm(np.cross(total_velocity_distr[i], wing.cp_dsl[i])) * G_list[i] \
                - Cl_i
            R_array[N_panels+i_glob] = R_array[i_glob]
            i_glob += 1
    return R_array


# @timeit
def calculate_corrector_equation(
    R_array: np.ndarray,
    total_velocity_dict: dict,
    aoa_eff_dict: dict,
    G_dict: dict,
    aoa_idx: int,
    wing_pool: WingPool,
    matrix_dim: int,
    linear_check: bool
) -> np.ndarray:
    """
    Newton corrector system of equations
    [J]delta_G = -R
    """
    J_matrix = np.zeros([matrix_dim, matrix_dim])
    i_glob = 0
    Cl_alpha_array = np.zeros(matrix_dim) # Used for debugging purposes
    for wing_i in wing_pool.complete_wing_pool:
        G_distr = G_dict[wing_i.surface_name]
        total_velocity_distr = total_velocity_dict[wing_i.surface_name]
        aoa_eff_distr = aoa_eff_dict[wing_i.surface_name]
        for i, _ in enumerate(wing_i.collocation_points):
            N_panels = wing_i.N_panels
            j_glob = 0
            w_i = np.cross(total_velocity_distr[i], wing_i.cp_dsl[i])
            w_i_abs = npla.norm(w_i)
            u_n_i = wing_i.u_n[i]
            u_a_i = wing_i.u_a[i]
            v_n_i = np.dot(total_velocity_distr[i], wing_i.u_n[i])
            v_a_i = np.dot(total_velocity_distr[i], wing_i.u_a[i])

            if linear_check:
                linear_data = get_linear_data_and_clmax(wing_i.cp_airfoils[i], wing_i.cp_reynolds[i], wing_i.airfoil_data)
                Cl_alpha_i = linear_data["cl_alpha"] * 180 / np.pi
            else:
                Cl_alpha_i = get_airfoil_data(
                        wing_i.cp_airfoils[i],
                        wing_i.cp_reynolds[i],
                        aoa_eff_distr[i] * 180 / np.pi,
                        wing_i.airfoil_data,
                        cl_alpha_check = True
                    ) * 180 / np.pi
            Cl_alpha_array[i_glob] = Cl_alpha_i
            for wing_j in wing_pool.complete_wing_pool:
                if "_mirrored" in wing_j.surface_name: 
                    j_glob += wing_j.N_panels
                    continue
                v_ij_distr = wing_pool.ind_velocities_list[aoa_idx][wing_i.surface_name][wing_j.surface_name]
                v_ij_distr_mirrored = wing_pool.ind_velocities_list[aoa_idx][wing_i.surface_name][wing_j.surface_name+"_mirrored"]
                for j, _ in enumerate(wing_j.collocation_points):
                    v_ij = v_ij_distr[i][j]
                    v_ij_m = v_ij_distr_mirrored[i][j]

                    coef_ij = 2 * np.dot(w_i,np.cross(v_ij, wing_i.cp_dsl[i]))*G_distr[i] / w_i_abs \
                        - Cl_alpha_i * (v_a_i * np.dot(v_ij, u_n_i) - v_n_i * np.dot(v_ij, u_a_i)) /(v_a_i ** 2 + v_n_i ** 2)
                    
                    coef_ij_m = 2 * np.dot(w_i,np.cross(v_ij_m, wing_i.cp_dsl[i]))*G_distr[i] / w_i_abs \
                        - Cl_alpha_i * (v_a_i * np.dot(v_ij_m, u_n_i) - v_n_i * np.dot(v_ij_m, u_a_i)) /(v_a_i ** 2 + v_n_i ** 2)
    
                    if wing_i.parent_wing == wing_j.surface_name and i_glob == (N_panels+j_glob):
                        coef_ij_m += 2 * w_i_abs

                    if wing_i.surface_name == wing_j.surface_name and i == j:
                        coef_ij += 2 * w_i_abs
                    
                    J_matrix[i_glob][j_glob] = coef_ij
                    J_matrix[i_glob][N_panels+j_glob] = coef_ij_m
                    j_glob += 1
            i_glob += 1

    delta_G = npla.solve(J_matrix, -R_array)
    return delta_G



from typing import List
import numpy as np
from dataclasses import dataclass, field


@dataclass(repr=False, eq=False, match_args=False, slots=True)
class FlightCondition:
    V_inf: float
    nu: float
    rho: float
    aoa: List[float]
    h: float = 0
    ground_effect_check: bool = False
    v_inf_list: np.ndarray = field(init=False)

    def __post_init__(self):
        self.v_inf_list = np.zeros((len(self.aoa), 3))
        for i, aoa_i in enumerate(self.aoa):
            aoa_i = aoa_i * np.pi / 180
            self.v_inf_list[i,:] = [np.cos(aoa_i), 0, np.sin(aoa_i)]



from dataclasses import dataclass, field
import numpy as np
from lltdrek.models.wingpool import WingPool
from lltdrek.utils.timeit import timeit
from lltdrek.simulation.main_equations import (
    calculate_corrector_equation,
    calculate_main_equation,
    calculate_main_equation_simplified
)


@dataclass(repr=False, eq=False, match_args=False, slots=True)
class Simulation:

    simulation_modes = [
        "linear_first",
        "latest_solution"
    ]

    wing_pool: WingPool
    damping_factor: float = 0.7
    max_iter: int = 150
    max_residual: float = 1e-3
    linear_check: bool = False
    show_logs: bool = True
    simulation_mode: str = "latest_solution"
    matrix_dim: int = field(init=False)


    def __post_init__(self):
        self.matrix_dim = sum([wing.N_panels for wing in self.wing_pool.complete_wing_pool])
        if self.simulation_mode not in self.simulation_modes:
            raise ValueError(f"Valor de simulation_mode ({self.simulation_mode}) inválido. Valores aceitos: {self.simulation_modes}")


    # @timeit
    def run_simulation(self) -> list:
        """
        Método para rodar a simulação principal do lltdrek
        A simulação usa o método da convergencia acelerada como default,
        onde o resultado da simulação anterior é o primeiro chute da simulação atual,
        acelerando a convergência
        """
        print(f"Running simulation for angles between {self.wing_pool.flight_condition.aoa[0]} and {self.wing_pool.flight_condition.aoa[-1]}") if self.show_logs is True else None
        print(f"Linear simulation check: {self.linear_check}") if self.show_logs is True else None
        G_solution_list = []
        for idx, aoa in enumerate(self.wing_pool.flight_condition.aoa):
            iteration = 1

            if idx == 0: 
                G = np.ones(self.matrix_dim) * 0.1
                G_dict = self.wing_pool.G_dict
            if self.simulation_mode == "linear_first":
                # Solve linear system to get a better approximation for G
                G = calculate_main_equation_simplified(
                    v_inf_array=self.wing_pool.flight_condition.v_inf_list[idx],
                    aoa_idx=idx,
                    wing_pool=self.wing_pool,
                    matrix_dim=self.matrix_dim
                    )
                G_linear_solution = G # debugging purposes
                G_dict = self.wing_pool.update_solution(G_solution=G)
                a=1

            total_velocity_dict = self.wing_pool.calculate_total_velocity(
                aoa_idx=idx,
                G_dict=G_dict
                )
            aoa_eff_dict = self.wing_pool.calculate_aoa_eff(total_velocity_dict)
            
            while True:
                R_array = calculate_main_equation(
                    total_velocity_dict,
                    aoa_eff_dict,
                    G_dict,
                    self.wing_pool,
                    self.matrix_dim,
                    self.linear_check
                )
                delta_G = calculate_corrector_equation(
                    R_array,
                    total_velocity_dict,
                    aoa_eff_dict,
                    G_dict,
                    idx,
                    self.wing_pool,
                    self.matrix_dim,
                    self.linear_check
                )

                if iteration > self.max_iter:
                    G_solution_list.append(np.nan)
                    if "last_successful_solution" in locals():
                        G_dict = last_successful_solution_dict
                    else:
                        G_dict = self.wing_pool.G_dict
                    print(f"Reached max iterations for angle {aoa}") if self.show_logs is True else None
                    break
                if abs(R_array.max()) < self.max_residual:
                    G_solution_list.append(G_dict)
                    print(f"Found solution for angle {aoa}") if self.show_logs is True else None
                    print(f"number of iterations: {iteration}") if self.show_logs is True else None
                    last_successful_solution_dict = G_dict
                    break
                else:
                    G = G + delta_G * self.damping_factor                  
                    G_dict = self.wing_pool.update_solution(G)

                    # Pré-calcular distribuição de alfas e velocidade total por painel
                    total_velocity_dict = self.wing_pool.calculate_total_velocity(
                        aoa_idx=idx,
                        G_dict=G_dict
                        )
                    aoa_eff_dict = self.wing_pool.calculate_aoa_eff(total_velocity_dict)
                    iteration += 1

        return G_solution_list



import numpy as np
import numpy.linalg as npla


def get_induced_velocity_distribution(
    collocation_points: np.ndarray,
    cp_macs: np.ndarray,
    vertice_points: np.ndarray,
    v_inf_array: np.ndarray,
    surface_name: str,
    ground_effect_check: bool,
    h: float
    ) -> np.ndarray:
    """
    Distribution of induced velocities
    """
    len_cp = len(collocation_points)
    len_vp = len(vertice_points)
    v_ij_distr = np.zeros([len_cp,len_vp-1,3])
    
    for i in range(len_cp):
        cp_i = collocation_points[i]
        mac_i = cp_macs[i]
        for j in range(len_vp - 1):
            if "_mirrored" in surface_name:
                # Pelo fato da asa ser espelhada, o sentido do vetor posição também é invertido
                vp_jj = vertice_points[j]
                vp_j = vertice_points[j+1]
            else:
                vp_j = vertice_points[j]
                vp_jj = vertice_points[j+1]
            v_ij, same_panel_check = get_induced_velocity(cp_i, vp_j, vp_jj, mac_i, v_inf_array)
            if ground_effect_check: 
                # O método do plano reflexivo inverte a asa, então vp_j e vp_jj precisam ser invertidos para que os sentidos dos vórtices estejam corretos
                v_ij_ge = get_induced_velocity_ground_effect(cp_i, vp_jj, vp_j, mac_i, v_inf_array, h, same_panel_check)
                v_ij += v_ij_ge
            v_ij_distr[i,j,:] = v_ij

    return v_ij_distr


def get_induced_velocity(
    collocation_point: np.ndarray,
    vertice_point_1: np.ndarray,
    vertice_point_2: np.ndarray,
    mac: np.ndarray,
    v_inf_array: np.ndarray
    ) -> np.ndarray: 
    """
    Function to calculate the induced velocity caused by a vortex panel in a point
    """
    velocity_ij = np.zeros(3)

    ri1j = collocation_point - vertice_point_1
    ri2j = collocation_point - vertice_point_2

    ri1j_abs = npla.norm(ri1j)
    ri2j_abs = npla.norm(ri2j)
    r1_cross_prod = np.cross(v_inf_array, ri1j)
    r2_cross_prod = np.cross(v_inf_array, ri2j)
    r1_dot_prod = np.dot(v_inf_array, ri1j)
    r2_dot_prod = np.dot(v_inf_array, ri2j)
    r12_cross_prod = np.cross(ri1j, ri2j)
    r12_dot_prod = np.dot(ri1j, ri2j)

    velocity_ij = mac / (4 * np.pi) * \
        ( r2_cross_prod / (ri2j_abs*(ri2j_abs-r2_dot_prod)) \
        - r1_cross_prod / (ri1j_abs*(ri1j_abs-r1_dot_prod)) )
    
    bound_vortex_den = (ri1j_abs*ri2j_abs*(ri1j_abs*ri2j_abs+r12_dot_prod))

    if bound_vortex_den != 0: # só adiciona essa parcela se a velocidade induzida não for em relação a si mesmo
        velocity_ij +=  mac / (4 * np.pi) * \
            ( (ri1j_abs+ri2j_abs) * r12_cross_prod / bound_vortex_den )
        same_panel_check = False
    else:
        same_panel_check = True
    
    return velocity_ij, same_panel_check


def get_induced_velocity_ground_effect(collocation_point: np.ndarray,
    vertice_point_1: np.ndarray,
    vertice_point_2: np.ndarray,
    mac: float,
    v_inf_array: np.ndarray,
    h: float,
    same_panel_check: bool 
    ) -> np.ndarray:
    """
    Induced velocity calculation considering ground effect through reflection method
    """
    velocity_ij = np.zeros(3)

    v_inf_ge = np.array(v_inf_array)
    v_inf_ge[2] = -v_inf_ge[2]

    ri1j = collocation_point - vertice_point_1
    ri1j[2] = collocation_point[2] - (2*h + 2*vertice_point_1[2])
    ri2j = collocation_point - vertice_point_2
    ri2j[2] = collocation_point[2] - (2*h + 2*vertice_point_2[2])

    ri1j_abs = npla.norm(ri1j)
    ri2j_abs = npla.norm(ri2j)
    r1_cross_prod = np.cross(v_inf_ge, ri1j)
    r2_cross_prod = np.cross(v_inf_ge, ri2j)
    r1_dot_prod = np.dot(v_inf_ge, ri1j)
    r2_dot_prod = np.dot(v_inf_ge, ri2j)
    r12_cross_prod = np.cross(ri1j, ri2j)
    r12_dot_prod = np.dot(ri1j, ri2j)

    bound_vortex_den = (ri1j_abs*ri2j_abs*(ri1j_abs*ri2j_abs+r12_dot_prod))

    velocity_ij = mac / (4 * np.pi) * \
        ( r2_cross_prod / (ri2j_abs*(ri2j_abs-r2_dot_prod)) \
        - r1_cross_prod / (ri1j_abs*(ri1j_abs-r1_dot_prod)) )
    
    if not same_panel_check: # só adiciona essa parcela se a velocidade induzida não for em relação a si mesmo
        velocity_ij +=  mac / (4 * np.pi) * \
            ( (ri1j_abs+ri2j_abs) * r12_cross_prod / bound_vortex_den )
    
    return velocity_ij



import os
import numpy as np
from typing import Union


AOA_POLYFIT_MIN = 0
AOA_POLYFIT_MAX = 8


def load_folder(folder_name: str, aoa_polyfit_min: Union[float, int] = None, aoa_polyfit_max: Union[float, int] = None):
    if aoa_polyfit_min is None or aoa_polyfit_max is None:
        aoa_polyfit_min = AOA_POLYFIT_MIN
        aoa_polyfit_max = AOA_POLYFIT_MAX
    # Get full directory to lookup files
    current_dir = os.path.abspath(os.getcwd())
    target_dir = os.path.join(current_dir, folder_name)
    files_list = []
    for (dir_path, _, filenames) in os.walk(target_dir):
        files_full_path = [(dir_path+"/"+file) for file in filenames]
        files_list.append(files_full_path)
    
    # Transform list of lists into a flat list
    files_flat_list = [item for sublist in files_list for item in sublist]
    txt_list = []
    dat_list = []
    for file in files_flat_list:
        file_extension = file.split(".")[-1]
        if file_extension == "txt": txt_list.append(file)
        elif file_extension == "dat": dat_list.append(file)
    
    airfoils_data_dict = {}
    for file in txt_list:
        airfoil_name = file.split("/")[-1].split(".")[0]
        airfoil_data = cl_file_to_dict(file, aoa_polyfit_min, aoa_polyfit_max)
        airfoils_data_dict[airfoil_name] = airfoil_data

    airfoils_dat_dict = {}
    for file in dat_list:
        airfoil_name = file.split("/")[-1].split(".")[0]
        dat_list = dat_file_to_dict(file)
        airfoils_dat_dict[airfoil_name] = dat_list
    
    return airfoils_data_dict, airfoils_dat_dict



def cl_file_to_dict(file_path: str, aoa_polyfit_min, aoa_polyfit_max):
    airfoil_data_dict = {}
    with open(file_path, "r") as f:
        while True:
            line = f.readline()
            if line == "": break
            reynolds = line.rstrip().split(",")[-1]
            cm0_line = f.readline()
            cm0 = float(cm0_line.rstrip().split(",")[-1])
            cl_list_temp = []
            while True:
                cl_line = f.readline()
                if cl_line == "\n" or cl_line == "": break
                else:
                    cl_line = cl_line.rstrip().split(",")
                    cl_list_temp.append([float(value) for value in cl_line])
            cl_linear_coefs = get_linear_coefs(cl_list_temp, aoa_polyfit_min, aoa_polyfit_max)
            cl_list = np.array(cl_list_temp)
            clmax = cl_list[:,1].max()
            airfoil_data_dict[reynolds] = {
                "cl_list": cl_list,
                "cl_alpha": cl_linear_coefs["cl_alpha"],
                "cl0": cl_linear_coefs["cl0"],
                "cm0": cm0,
                "clmax": clmax
                }
    return airfoil_data_dict


def dat_file_to_dict(file_path: str):
    dat_list = []
    dat_file = np.loadtxt(file_path,skiprows=1)
    for row in dat_file:
        dat_list.append([coordinate for coordinate in row])
    return dat_list


def get_linear_coefs(cl_list: list, aoa_polyfit_min, aoa_polyfit_max):
    aoa_interp_list = []
    cl_interp_list = []
    for values in cl_list:
        if aoa_polyfit_min <= values[0] <= aoa_polyfit_max:
            aoa_interp_list.append(values[0])
            cl_interp_list.append(values[1])
    linear_coefs = np.polyfit(aoa_interp_list, cl_interp_list,1)
    cl_linear_coefs_dict = {
        "cl_alpha": linear_coefs[0],
        "cl0": linear_coefs[1]
    }
    return cl_linear_coefs_dict
        

def xfoil_file_to_cl_file(file_path: str):
    # TODO: implementar
    pass



from functools import wraps
import time


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__} Took {total_time:.8f} seconds')
        return result
    return timeit_wrapper


from typing import List, Union
import numpy as np
from dataclasses import dataclass, field
from lltdrek.models.wingpool import WingPool
from lltdrek.models.wing import Wing
from lltdrek.models.flight_condition import FlightCondition
from lltdrek.utils.lookup import get_airfoil_data, get_linear_data_and_clmax


@dataclass(repr=False, eq=False, match_args=False, slots=True)
class PostProcessing:
    ref_point: list
    ref_points_dict: dict = field(init=False)


    def __post_init__(self) -> None:
        self.ref_points_dict = None

    def get_global_coefficients(self, wing_pool: WingPool, G_dict: dict[list], aoa_index: int, S_ref: float, c_ref: float) -> dict:
        if self.ref_points_dict is None:
            self.build_reference_points_dict(wing_pool)
        
        if type(G_dict) != dict:
            CF = np.array((np.nan, np.nan, np.nan))
            CM = np.array((np.nan, np.nan, np.nan))
            Cl_distr_dict = {}
            for wing in wing_pool.complete_wing_pool:
                Cl_distr_dict[wing.surface_name] = np.nan
            return {
                "CF": CF,
                "CM": CM,
                "Cl_distr": Cl_distr_dict
            }

        v_inf = wing_pool.flight_condition.v_inf_list[aoa_index]
        aoa = wing_pool.flight_condition.aoa[aoa_index]
        aoa_rad = aoa * np.pi / 180

        CF = np.zeros(3)
        CM = np.zeros(3)
        CF_distr = np.zeros((wing_pool.total_panels, 3))
        CM_distr = np.zeros((wing_pool.total_panels, 3))
        i_glob = 0
        Cl_distr_dict = {}
        for wing_i in wing_pool.complete_wing_pool:
            ref_points_distr = self.ref_points_dict[wing_i.surface_name]
            Cl_distr = np.zeros(wing_i.N_panels)
            G_i = G_dict[wing_i.surface_name]
            for i, _ in enumerate(wing_i.collocation_points):
                aux_cf = np.zeros(3)

                airfoil_coefficients = get_linear_data_and_clmax(
                    wing_i.cp_airfoils[i],
                    wing_i.cp_reynolds[i],
                    wing_i.airfoil_data
                )
                cm_i = airfoil_coefficients["cm0"]
                Cl_distr[i] = 2 * G_i[i]
                for wing_j in wing_pool.complete_wing_pool:
                    G_j = G_dict[wing_j.surface_name]
                    v_ij_distr = wing_pool.ind_velocities_list[aoa_index][wing_i.surface_name][wing_j.surface_name]
                    for j, _ in enumerate(wing_j.collocation_points):
                        v_ij = v_ij_distr[i][j]
                        aux_cf += G_j[j] * v_ij
                aux_cf = (aux_cf + v_inf) * G_i[i]
                CF += 2 * np.cross(aux_cf, wing_i.cp_dsl[i]) * wing_i.cp_areas[i] / S_ref
                CF_distr[i_glob,:] = 2 * np.cross(aux_cf, wing_i.cp_dsl[i]) * wing_i.cp_areas[i] / S_ref
                CM += (2 * np.cross(ref_points_distr[i], np.cross(aux_cf, wing_i.cp_dsl[i])) - cm_i * wing_i.cp_macs[i] * wing_i.u_s[i]) * wing_i.cp_areas[i] / ( S_ref * c_ref )
                CM_distr[i_glob,:] = (2 * np.cross(ref_points_distr[i], np.cross(aux_cf, wing_i.cp_dsl[i])) - cm_i * wing_i.cp_macs[i] * wing_i.u_s[i]) * wing_i.cp_areas[i] / ( S_ref * c_ref )

                i_glob += 1
            Cl_distr_dict[wing_i.surface_name] = Cl_distr

        # Rebater os coeficientes de forças para o eixo do escoamento
        rotation_matrix = np.array(
            [[np.cos(aoa_rad), 0, np.sin(aoa_rad)],
            [0, 1, 0],
            [-1*np.sin(aoa_rad), 0, np.cos(aoa_rad)]]
        )
        CF = np.matmul(rotation_matrix, CF)
        return {
            "CF": CF,
            "CM": CM,
            "Cl_distr": Cl_distr_dict
        }

    
    def build_reference_points_dict(self, wing_pool: WingPool) -> dict:
        self.ref_points_dict = {}
        for wing_i in wing_pool.wing_list:
            ref_point_distr = np.zeros((wing_i.N_panels, 3))
            ref_point_distr_mirrored  = np.zeros((wing_i.N_panels, 3))
            for i, cp_i in enumerate(wing_i.collocation_points):
                ref_point_distr[i,:] = cp_i - self.ref_point
                ref_point_distr_mirrored[i,:] = cp_i - self.ref_point
                ref_point_distr_mirrored[i,1] = ref_point_distr_mirrored[i,1] * -1

            self.ref_points_dict[wing_i.surface_name] = ref_point_distr
            self.ref_points_dict[wing_i.surface_name+"_mirrored"] = ref_point_distr_mirrored
        
        return self.ref_points_dict


    def get_wing_coefficients(self, wing_pool: WingPool, G_dict: dict[list], aoa_index: int, S_ref: float, c_ref: float) -> dict:
        if self.ref_points_dict is None:
            self.build_reference_points_dict(wing_pool)
        
        wing_coefficients = {}

        if type(G_dict) != dict:
            CF = np.array((np.nan, np.nan, np.nan))
            CM = np.array((np.nan, np.nan, np.nan))
            Cl_distr_dict = {}
            for wing in wing_pool.complete_wing_pool:
                Cl_distr_dict[wing.surface_name] = np.nan
                wing_coefficients[wing.surface_name] = {
                    "CF": CF,
                    "CM": CM,
                    "Cl_distr": Cl_distr_dict
                }
            return wing_coefficients
        
        v_inf = wing_pool.flight_condition.v_inf_list[aoa_index]
        aoa = wing_pool.flight_condition.aoa[aoa_index]
        aoa_rad = aoa * np.pi / 180

        CF_distr = np.zeros((wing_pool.total_panels, 3))
        CM_distr = np.zeros((wing_pool.total_panels, 3))
        i_glob = 0
        Cl_distr_dict = {}
        for wing_i in wing_pool.complete_wing_pool:
            CF = np.zeros(3)
            CM = np.zeros(3)
            
            ref_points_distr = self.ref_points_dict[wing_i.surface_name]
            Cl_distr = np.zeros(wing_i.N_panels)
            G_i = G_dict[wing_i.surface_name]
            for i, _ in enumerate(wing_i.collocation_points):
                aux_cf = np.zeros(3)

                airfoil_coefficients = get_linear_data_and_clmax(
                    wing_i.cp_airfoils[i],
                    wing_i.cp_reynolds[i],
                    wing_i.airfoil_data
                )
                cm_i = airfoil_coefficients["cm0"]
                Cl_distr[i] = 2 * G_i[i]
                for wing_j in wing_pool.complete_wing_pool:
                    G_j = G_dict[wing_j.surface_name]
                    v_ij_distr = wing_pool.ind_velocities_list[aoa_index][wing_i.surface_name][wing_j.surface_name]
                    for j, _ in enumerate(wing_j.collocation_points):
                        v_ij = v_ij_distr[i][j]
                        aux_cf += G_j[j] * v_ij
                aux_cf = (aux_cf + v_inf) * G_i[i]
                CF += 2 * np.cross(aux_cf, wing_i.cp_dsl[i]) * wing_i.cp_areas[i] / S_ref
                CF_distr[i_glob,:] = 2 * np.cross(aux_cf, wing_i.cp_dsl[i]) * wing_i.cp_areas[i] / S_ref
                CM += (2 * np.cross(ref_points_distr[i], np.cross(aux_cf, wing_i.cp_dsl[i])) - cm_i * wing_i.cp_macs[i] * wing_i.u_s[i]) * wing_i.cp_areas[i] / ( S_ref * c_ref )
                CM_distr[i_glob,:] = (2 * np.cross(ref_points_distr[i], np.cross(aux_cf, wing_i.cp_dsl[i])) - cm_i * wing_i.cp_macs[i] * wing_i.u_s[i]) * wing_i.cp_areas[i] / ( S_ref * c_ref )
                i_glob += 1

            rotation_matrix = np.array(
                [[np.cos(aoa_rad), 0, np.sin(aoa_rad)],
                [0, 1, 0],
                [-1*np.sin(aoa_rad), 0, np.cos(aoa_rad)]]
            )
            CF = np.matmul(rotation_matrix, CF)
            Cl_distr_dict[wing_i.surface_name] = Cl_distr
            wing_coefficients[wing_i.surface_name] = {
                "CF": CF,
                "CM": CM,
                "Cl_distr": Cl_distr_dict
            }

        for wing_i in wing_pool.wing_list:
            # Somar as parcelas das asas (wing + wing_mirrored) e remover chaves de asas espelhadas
            wing_coefficients[wing_i.surface_name]["CF"] += wing_coefficients[wing_i.surface_name+"_mirrored"]["CF"]
            wing_coefficients[wing_i.surface_name]["CM"] += wing_coefficients[wing_i.surface_name+"_mirrored"]["CM"]
            wing_coefficients.pop(wing_i.surface_name+"_mirrored")

        return wing_coefficients


    def get_CL_max_linear(
        self,
        wing_pool: WingPool,
        G_list: list[dict],
        aoa_range: Union[tuple[float],None],
        S_ref: float,
    ) -> float:
        """
        Obtenção do CL máximo de uma asa (ou sistema de asas) através do método da seção crítica
        Para cada ângulo é verificado o Cl de cada seção e comparado com o Clmax do perfil no seu respectivo reynolds
        O CLmax é alcançado quando alguma seção atingir o Clmax do seu perfil no seu respectivo reynolds
        O input aoa_index_range pode ser utilizado para procurar numa faixa de angulos específica
        por exemplo, se aoa_list [1, 3, 5, 7, 9, 11, 13, 15, 19]
        passando-se aoa_range = [13, 19] será iterado somente entre os ângulos 13 e 19
        """
        if aoa_range == None:
            aoa_start = wing_pool.flight_condition.aoa[0]
            aoa_end = wing_pool.flight_condition.aoa[-1]
        else:
            aoa_start = aoa_range[0]
            aoa_end = aoa_range[-1]

        CLmax_check = False
        for aoa_idx, aoa in enumerate(wing_pool.flight_condition.aoa):
            if aoa_start <= aoa <= aoa_end and not CLmax_check:
                G_dict = G_list[aoa_idx]
                if type(G_dict) != dict: continue # Pula a iteração caso G_dict = nan (não convergiu)
                for wing_i in wing_pool.wing_list:
                    G_i = G_dict[wing_i.surface_name]
                    for i, _ in enumerate(wing_i.collocation_points):
                        lookup_data = get_linear_data_and_clmax(
                            wing_i.cp_airfoils[i],
                            wing_i.cp_reynolds[i],
                            wing_i.airfoil_data
                        )
                        Clmax_i = lookup_data["clmax"]
                        if 2 * G_i[i] > Clmax_i:
                            CLmax_check = True
                            aoa_max_idx = aoa_idx - 1 if aoa_idx > 0 else 0
        if not CLmax_check:
            aoa_max_idx = -1 if aoa_range == None else wing_pool.flight_condition.aoa.index(aoa_range[-1])
        G_dict = G_list[aoa_max_idx]
        aoa_max = wing_pool.flight_condition.aoa[aoa_max_idx]
        c_ref = 1 # O valor de c_ref pode ser qualquer um aqui porque nao vamos pegar coeficiente de momento
        coefficients = self.get_global_coefficients(wing_pool, G_dict, aoa_max_idx, S_ref, c_ref)
        return {
            "CLmax": coefficients["CF"][2],
            "aoa_max": aoa_max,
            "aoa_max_idx": aoa_max_idx
        }


    def get_aerodynamic_center(self, wing_pool: WingPool, G_list: list[dict], aoa_1: float, aoa_2: float, S_ref: float, c_ref: float) -> dict:
        aoa_list = wing_pool.flight_condition.aoa
        if aoa_1 not in aoa_list or aoa_2 not in aoa_list:
            raise Exception(f"Ambos aoa_1 e aoa_2 precisam estar dentro da lista de ângulos de ataque")

        aoa_1_idx = aoa_list.index(aoa_1)
        aoa_2_idx = aoa_list.index(aoa_2)

        ac = 0.25*c_ref
        ac_min = 0.1*c_ref
        ac_max = 1.5*c_ref
        max_iter = 100
        
        self.ref_point = [ac, 0, 0]
        ac_check = False
        
        i = 1
        while not ac_check:
            if i > max_iter:
                print("Limite máximo de iterações atingido")
                break
            
            self.build_reference_points_dict(wing_pool)

            coefs_1 = self.get_global_coefficients(wing_pool, G_list[0], aoa_index=aoa_1_idx, S_ref=S_ref, c_ref=c_ref)
            Cm_1 = coefs_1["CM"][1]
            coefs_2 = self.get_global_coefficients(wing_pool, G_list[1], aoa_index=aoa_2_idx, S_ref=S_ref, c_ref=c_ref)
            Cm_2 = coefs_2["CM"][1]

            Cm_alpha = (Cm_2 - Cm_1) / (aoa_2 - aoa_1)
            if abs(Cm_alpha) <= 1e-6 or i == max_iter:
                Cm_ac = Cm_1
                ac_check = True
                x_ac = ac / c_ref
            else:
                if Cm_alpha < 0:
                    ac_min = ac
                    ac = (ac + ac_max) / 2
                    self.ref_point = [ac, 0, 0]
                else:
                    ac_max = ac
                    ac = (ac_min + ac_max) / 2
                    self.ref_point = [ac, 0, 0]
            i += 1

        return {
            "x_ac": x_ac,
            "Cm_ac": Cm_ac
        }
